from .splits import *
from .config import args_object_from_args_dict
from .train import *
from .predict import *
from .interpret import * 
import argparse
import pdb

def parse_args():
    parser=argparse.ArgumentParser()
    parser.add_argument("--assembly",default="hg19")
    parser.add_argument("--data_path",help="path that stores training/validation/test data")
    parser.add_argument("--model_hdf5",required=True)
    parser.add_argument("--batch_size",type=int,default=1000)
    parser.add_argument("--init_weights",default=None)
    parser.add_argument("--ref_fasta",default="/mnt/data/annotations/by_release/hg19.GRCh37/hg19.genome.fa")
    parser.add_argument("--w0_file",default=None)
    parser.add_argument("--w1_file",default=None)
    parser.add_argument("--weighted",action="store_true")
    parser.add_argument("--from_checkpoint_weights",default=None)
    parser.add_argument("--from_checkpoint_arch",default=None)
    parser.add_argument("--num_tasks",required=True,type=int)
    parser.add_argument("--num_train",type=int,default=700000)
    parser.add_argument("--num_valid",type=int,default=150000)

    #add functionality to train on individuals' allele frequencies
    parser.add_argument("--vcf_file",default=None)
    parser.add_argument("--global_vcf",action="store_true")
    parser.add_argument("--revcomp",action="store_true")
    parser.add_argument("--epochs",type=int,default=40)
    parser.add_argument("--patience",type=int,default=3)
    parser.add_argument("--patience_lr",type=int,default=2,help="number of epochs with no drop in validation loss after which to reduce lr")
    parser.add_argument("--architecture_spec",type=str,default="basset_architecture_multitask")
    parser.add_argument("--architecture_from_file",type=str,default=None)
    parser.add_argument("--tensorboard",action="store_true")
    parser.add_argument("--tensorboard_logdir",default="logs")
    parser.add_argument("--squeeze_input_for_gru",action="store_true")
    parser.add_argument("--seed",type=int,default=1234)
    parser.add_argument("--train_upsample", type=float, default=None)
    parser.add_argument("--valid_upsample", type=float, default=None)
    parser.add_argument("--threads",type=int,default=1)
    parser.add_argument("--max_queue_size",type=int,default=100)
    parser.add_argument("--save_w1_w0", default=None,help="output text file to save w1 and w0 to")
    parser.add_argument('--w1',nargs="*", type=float, default=None)
    parser.add_argument('--w0',nargs="*", type=float, default=None)
    parser.add_argument('--weights',help='weights file for the model')
    parser.add_argument('--yaml',help='yaml file for the model')
    parser.add_argument('--json',help='json file for the model')
    parser.add_argument('--predict_chroms',default=None) 
    parser.add_argument('--data_hammock',help='input file is in hammock format, with unique id for each peak')
    parser.add_argument('--variant_bed')
    parser.add_argument('--predictions_pickle',help='name of pickle to save predictions',default=None)
    parser.add_argument('--accuracy_metrics_file',help='file name to save accuracy metrics',default=None)
    parser.add_argument('--predictions_pickle_to_load',help="if predictions have already been generated, provide a pickle with them to just compute the accuracy metrics",default=None)
    parser.add_argument('--background_freqs',default=None)
    parser.add_argument('--flank',default=500,type=int)
    parser.add_argument('--mask',default=10,type=int)
    parser.add_argument('--center_on_summit',default=False,action='store_true',help="if this is set to true, the peak will be centered at the summit (must be last entry in bed file or hammock) and expanded args.flank to the left and right")
    parser.add_argument("--interpret_chroms",nargs="*") 
    parser.add_argument("--interpretation_outf",default=None)
    parser.add_argument("--method",choices=['gradxinput','deeplift'],default="deeplift")
    parser.add_argument('--task_id',type=int)
    parser.add_argument('--chromsizes',default='/mnt/data/annotations/by_release/hg19.GRCh37/hg19.chrom.sizes')    
    return parser.parse_args()


def cross_validate(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args) 

    #run training on each of the splits
    if args.assembly not in splits:
        raise Exception("Unsupported genome assembly:"+args.assembly+". Supported assemblies include:"+str(splits.keys())+"; add splits for this assembly to splits.py file")
    args_dict=vars(args)
    print(args_dict) 
    base_model_file=str(args_dict['model_hdf5'])
    base_accuracy_file=str(args_dict['accuracy_metrics_file'])
    base_interpretation=str(args_dict['interpretation_outf'])


    for split in splits[args.assembly]:
        print("Starting split:"+str(split))
        test_chroms=splits[args.assembly][split]['test']
        validation_chroms=splits[args.assembly][split]['valid']
        train_chroms=list(set(chroms[args.assembly])-set(test_chroms+validation_chroms))

        #convert args to dict
        args_dict=vars(args)
        args_dict['train_chroms']=train_chroms
        args_dict['validation_chroms']=validation_chroms
        
           
        #set the training arguments specific to this fold 
        args_dict['model_hdf5']=base_model_file+"."+str(split)
        print("Training model") 
        train(args_dict)

        #set the prediction arguments specific to this fold
        args_dict['accuracy_metric_file']=base_accuracy_file+"."+str(split)
        args_dict['predict_chroms']=test_chroms
        print("Calculating predictions on the test fold") 
        predict(args_dict)
        
        args_dict['interpret_chroms']=test_chroms
        args_dict['interpretation_outf']=base_interpretation+'.'+str(split)
        print("Running interpretation on the test fold") 
        interpret(args_dict)
        
        
def main():
    args=parse_args()
    cross_validate(args) 

if __name__=="__main__":
    main()



import keras.backend as K
def get_weighted_binary_crossentropy(w0_weights, w1_weights):
    # Compute the task-weighted cross-entropy loss, where every task is weighted by 1 - (fraction of non-ambiguous examples that are positive)
    # In addition, weight everything with label -1 to 0
    import numpy as np 
    w0_weights=np.array(w0_weights);
    w1_weights=np.array(w1_weights);
    thresh=-0.5
    def weighted_binary_crossentropy(y_true,y_pred):
        weightsPerTaskRep = y_true*w1_weights[None,:] + (1-y_true)*w0_weights[None,:]
        nonAmbig = K.cast((y_true > -0.5),'float32')
        nonAmbigTimesWeightsPerTask = nonAmbig * weightsPerTaskRep
        return K.mean(K.binary_crossentropy(y_pred, y_true)*nonAmbigTimesWeightsPerTask, axis=-1);
    return weighted_binary_crossentropy; 





#compute gradient x input for tensorflow models.
import argparse
import pdb 
from deeplift.conversion import kerasapi_conversion as kc
import numpy as np
import pyBigWig
from .config import args_object_from_args_dict

def parse_args():
    parser=argparse.ArgumentParser(description="get gradient x input for a model")
    parser.add_argument("--model_hdf5")
    parser.add_argument("--w0",nargs="+",type=float)
    parser.add_argument("--w1",nargs="+",type=float)
    parser.add_argument("--data_path")
    parser.add_argument("--interpret_chroms",nargs="*") 
    parser.add_argument("--interpretation_outf",default=None)
    parser.add_argument("--flank",default=None,type=int)
    parser.add_argument("--method",choices=['gradxinput','deeplift'],default="deeplift")
    parser.add_argument('--batch_size',type=int,help='batch size to use to make model predictions',default=50)
    parser.add_argument('--ref_fasta',default="/srv/scratch/annashch/deeplearning/form_inputs/code/hg19.genome.fa")
    parser.add_argument('--background_freqs',default=None)
    parser.add_argument('--center_on_summit',default=False,action='store_true',help="if this is set to true, the peak will be centered at the summit (must be last entry in bed file or hammock) and expanded args.flank to the left and right")
    parser.add_argument('--task_id',type=int)
    parser.add_argument('--squeeze_input_for_gru',default=False,action='store_true')
    parser.add_argument('--assembly',default='hg19')
    parser.add_argument('--chromsizes',default='/mnt/data/annotations/by_release/hg19.GRCh37/hg19.chrom.sizes')
    return parser.parse_args()


def get_deeplift_function(args):
    # convert to deeplift model and get scoring function
    deeplift_model = kc.convert_model_from_saved_files(args.model_hdf5,verbose=False)

    #get the deeplift score with respect to the logit 
    score_func = deeplift_model.get_target_contribs_func(
        find_scores_layer_idx=0,
        target_layer_idx=-2)
    return score_func

def get_deeplift_references(args):
    if args.background_freqs==None:
        # use a 40% GC reference
        input_references = [np.array([0.3, 0.2, 0.2, 0.3])[None, None, None, :]]
    else:
        input_references=[np.array(args.background_freqs)[None,None,None,:]]
    return input_references

def add_bigwig_header(bw,assembly):
    if assembly=='hg19':
        bw.addHeader([('chr1',249250621),
                      ('chr2',243199373),                      
                      ('chr3',198022430),
                      ('chr4',191154276),
                      ('chr5',180915260),
                      ('chr6',171115067),
                      ('chr7',159138663),
                      ('chr8',146364022),
                      ('chr9',141213431),
                      ('chr10',135534747),
                      ('chr11',135006516),
                      ('chr12',133851895),
                      ('chr13',115169878),
                      ('chr14',107349540),
                      ('chr15',102531392),
                      ('chr16',90354753),
                      ('chr17',81195210),
                      ('chr18',78077248),
                      ('chr19',59128983),
                      ('chr20',63025520),
                      ('chr21',48129895),
                      ('chr22',51304566),
                      ('chrX',155270560),
                      ('chrY',59373566)])
        return bw
    else:
        raise Exception("implement bigWig header for this assembly!")
def get_chromsizes(f):
    data=open(f,'r').read().strip().split('\n')
    chromsize_dict=dict()
    for line in data:
        tokens=line.split()
        chromsize_dict[tokens[0]]=int(tokens[1])
    return chromsize_dict

def get_deeplift_scores_bed(args,score_func,input_references):
    import pysam
    import pandas as pd
    num_generated=0
    ref=pysam.FastaFile(args.ref_fasta) 
    data=pd.read_csv(args.data_path,header=None,sep='\t')
    ltrdict = {'a':[1,0,0,0],'c':[0,1,0,0],'g':[0,0,1,0],'t':[0,0,0,1], 'n':[0,0,0,0],'A':[1,0,0,0],'C':[0,1,0,0],'G':[0,0,1,0],'T':[0,0,0,1],'N':[0,0,0,0]}
    #iterate through batches and one-hot-encode on the fly
    num_entries=data.shape[0]
    bw=pyBigWig.open(args.interpretation_outf,'w')
    bw=add_bigwig_header(bw,args.assembly)
    chromsize_dict=get_chromsizes(args.chromsizes)
    while num_generated < num_entries:
        print(str(num_generated))
        start_index=num_generated
        end_index=min([num_entries,start_index+args.batch_size])
        seqs=[]
        chroms=[]
        start_vals=[]
        end_vals=[]
        for i in range(start_index,end_index):
            cur_row=data.iloc[i]
            chrom=cur_row[0]
            start_val=cur_row[1]
            end_val=cur_row[2]
            if args.center_on_summit==True:
                summit_offset=int(cur_row[-1])
                summit_pos=start_val+summit_offset
                start_val=summit_pos - args.flank
                end_val=summit_pos+args.flank
            if start_val<1:
                start_val=1
                end_val=1+2*args.flank
            if end_val>chromsize_dict[chrom]:
                end_val=chromsize_dict[chrom]-1
                start_val=end_val-2*args.flank 
            try:
                seq=ref.fetch(chrom,start_val,end_val)
                seqs.append(seq)
                chroms.append(chrom)
                start_vals.append(start_val)
                end_vals.append(end_val) 
            except:
                continue
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        if (args.squeeze_input_for_gru==False):
            #expand dimension of 1
            x=np.expand_dims(seqs,1)
        else:
            x=seqs

        cur_scores = score_func(
            task_idx=args.task_id,
            input_data_list=[x],
            batch_size=x.shape[0],
            progress_update=None,
            input_references_list=input_references)*x        
        for i in range(len(cur_scores)):
            base_scores=np.ndarray.tolist(np.sum(cur_scores[i].squeeze(),axis=1))
            try:
                bw.addEntries(chroms[i],start_vals[i],values=base_scores,span=1,step=1)
            except:
                continue
        num_generated+=(end_index-start_index)
    bw.close()
    return 
def get_gradxinput_scores_bed(args,normed_grad):
    import pysam
    import pandas as pd
    num_generated=0
    ref=pysam.FastaFile(args.ref_fasta) 
    data=pd.read_csv(args.data_hammock,header=None,sep='\t')
    ltrdict = {'a':[1,0,0,0],'c':[0,1,0,0],'g':[0,0,1,0],'t':[0,0,0,1], 'n':[0,0,0,0],'A':[1,0,0,0],'C':[0,1,0,0],'G':[0,0,1,0],'T':[0,0,0,1],'N':[0,0,0,0]}
    #iterate through batches and one-hot-encode on the fly
    num_entries=data.shape[0]
    outf=open(args.interpretation_outf,'w')
    while num_generated < num_entries:
        print(str(num_generated))
        start_index=num_generated
        end_index=min([num_entries,start_index+args.batch_size])
        seqs=[]
        chroms=[]
        start_vals=[]
        end_vals=[]
        for i in range(start_index,end_index):
            cur_row=data.iloc[i]
            chrom=cur_row[0]
            start_val=cur_row[1]
            end_val=cur_row[2]
            if args.center_on_summit==True:
                summit_offset=int(cur_row[-1])
                summit_pos=start_val+summit_offset
                start_val=summit_pos - args.flank
                end_val=summit_pos+args.flank
                if start_val<1:
                    start_val=1
                    end_val=1+2*args.flank 
            try:
                seq=ref.fetch(chrom,start_val,end_val)
                seqs.append(seq)
                chroms.append(chrom)
                start_vals.append(start_val)
                end_vals.append(end_val) 
            except:
                continue
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        if (args.squeeze_input_for_gru==False):
            #expand dimension of 1
            x=np.expand_dims(seqs,1)
        else:
            x=seqs
        cur_scores = normed_grad*x
        pdb.set_trace()
        for i in range(len(cur_scores)):
            entry=[chroms[i],start_vals[i],end_vals[i],cur_scores[i]]
            outf.write('\t'.join([str(j) for j in entry])+'\n')
        num_generated+=(end_index-start_index)


def interpret(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args)
    if args.method=="deeplift":
        # get deeplift scores
        score_func=get_deeplift_function(args)
        input_references=get_deeplift_references(args)
        get_deeplift_scores_bed(args,score_func,input_references)
        
    elif args.method=="gradxinput":
        #calculate gradient x input 
        model=get_model(args)
        import tensorflow as tf
        grad_tensor=K.gradients(model.layers[-2].output,model.layers[0].input)
        grad_func = K.function([model.layers[0].input,K.learning_phase()], grad_tensor)
        gradient = grad_func([inputs, False])[0]
        normed_gradient = gradient-np.mean(gradient, axis=3)[:,:,:,None]
        get_gradxinput_scores_bed(args,normed_grad)
        normed_grad_times_inp = normed_gradient*inputs
    else:
        raise Exception("method must be one of 'deeplift' or 'gradxinput'")
        
def main():
    args=parse_args()
    interpret(args) 
    
if __name__=="__main__":
    main()
    



background_freqs=[0.277,0.277,0.223,0.223]
ref="/srv/scratch/annashch/deeplearning/form_inputs/code/hg19.genome.fa"
mask=10
flank=500
variant_bed="/srv/scratch/annashch/gecco/variants.bed"
###
import pysam
import pdb
import random
import numpy as np 
ref=pysam.FastaFile(ref)
data=[i.split('\t') for i in open(variant_bed,'r').read().strip().split('\n')]
ltrdict = {'a':[1,0,0,0],'c':[0,1,0,0],'g':[0,0,1,0],'t':[0,0,0,1], 'n':[0,0,0,0],'A':[1,0,0,0],'C':[0,1,0,0],'G':[0,0,1,0],'T':[0,0,0,1],'N':[0,0,0,0]}
#original
seqs=[]
#snp
seqs_snp=[]
#n bases around snp (i.e. knock out an enhancer)
seqs_enhancer=[] 
for entry in data[1::]:
    #introduce the specified variant 
    start_pos=int(entry[1])-(flank)
    end_pos=int(entry[1])+flank
    seq=ref.fetch(entry[0],start_pos,end_pos)
    seqs.append(seq)
    alt_allele=entry[3]
    if alt_allele=="NA":
        #randomly insert another base
        ref_allele=seq[flank-1]
        options=['A','C','G','T']
        options.remove(ref_allele)
        alt_allele=options[random.randint(0,2)]
    seq=seq[0:flank-1]+alt_allele+seq[flank:len(seq)]
    seqs_snp.append(seq)
    seq=np.array([ltrdict[x] for x in seq])
    start_mask=flank-1-mask
    end_mask=flank-1+mask
    seq=seq*1.0
    seq[start_mask:end_mask]=background_freqs
    
    pdb.set_trace()
    



#Raw keras model
import importlib
import imp
import argparse
import numpy as np
import h5py
from .generators import *
from .config import args_object_from_args_dict
import pdb

def parse_args():
    parser=argparse.ArgumentParser()
    parser.add_argument("--data_path")
    parser.add_argument("--train_chroms",nargs="*",default=None)
    parser.add_argument("--validation_chroms",nargs="*",default=None) 
    parser.add_argument("--train_path")
    parser.add_argument("--valid_path")
    parser.add_argument("--model_hdf5")
    parser.add_argument("--batch_size",type=int,default=1000)
    parser.add_argument("--init_weights",default=None)
    parser.add_argument("--num_train",type=int,default=700000)
    parser.add_argument("--num_valid",type=int,default=150000)
    parser.add_argument("--ref_fasta",default="/mnt/data/annotations/by_release/hg19.GRCh37/hg19.genome.fa")
    parser.add_argument("--w0_file",default=None)
    parser.add_argument("--w1_file",default=None)
    parser.add_argument("--weighted",action="store_true")
    parser.add_argument("--from_checkpoint_weights",default=None)
    parser.add_argument("--from_checkpoint_arch",default=None)
    parser.add_argument("--num_tasks",type=int)
    #add functionality to train on individuals' allele frequencies
    parser.add_argument("--vcf_file",default=None)
    parser.add_argument("--global_vcf",action="store_true")
    parser.add_argument("--revcomp",action="store_true")
    parser.add_argument("--epochs",type=int,default=40)
    parser.add_argument("--patience",type=int,default=3)
    parser.add_argument("--patience_lr",type=int,default=2,help="number of epochs with no drop in validation loss after which to reduce lr")
    parser.add_argument("--architecture_spec",type=str,default="basset_architecture_multitask")
    parser.add_argument("--architecture_from_file",type=str,default=None)
    parser.add_argument("--tensorboard",action="store_true")
    parser.add_argument("--tensorboard_logdir",default="logs")
    parser.add_argument("--squeeze_input_for_gru",action="store_true")
    parser.add_argument("--seed",type=int,default=1234)
    parser.add_argument("--train_upsample", type=float, default=None)
    parser.add_argument("--valid_upsample", type=float, default=None)
    parser.add_argument("--threads",type=int,default=1)
    parser.add_argument("--max_queue_size",type=int,default=100)
    parser.add_argument("--save_w1_w0", default=None,help="output text file to save w1 and w0 to")
    parser.add_argument('--w1',nargs="*", type=float, default=None)
    parser.add_argument('--w0',nargs="*", type=float, default=None)
    return parser.parse_args()

    

def fit_and_evaluate(model,train_gen,valid_gen,args):
    model_output_path = args.model_hdf5
    from keras.callbacks import ModelCheckpoint, EarlyStopping, CSVLogger, ReduceLROnPlateau
    checkpointer = ModelCheckpoint(filepath=model_output_path, verbose=1, save_best_only=True)
    earlystopper = EarlyStopping(monitor='val_loss', patience=args.patience, verbose=1)
    csvlogger = CSVLogger(args.model_hdf5+".log", append = True)
    reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.4,patience=args.patience_lr, min_lr=0.00000001)
    cur_callbacks=[checkpointer,earlystopper,csvlogger,reduce_lr]
    if args.tensorboard==True:
        from keras.callbacks import TensorBoard
        #create the specified logdir
        import os
        cur_logdir='/'.join([args.tensorboard_logdir,args.model_hdf5+'.tb'])
        if not os.path.exists(cur_logdir):
                os.makedirs(cur_logdir)
        tensorboard_visualizer=TensorBoard(log_dir=cur_logdir, histogram_freq=0, batch_size=500, write_graph=True, write_grads=False, write_images=False, embeddings_freq=0, embeddings_layer_names=None, embeddings_metadata=None)
        cur_callbacks.append(tensorboard_visualizer)
    np.random.seed(1234)
    model.fit_generator(train_gen,
                        validation_data=valid_gen,
                        steps_per_epoch=args.num_train/args.batch_size,
                        validation_steps=args.num_valid/args.batch_size,
                        epochs=args.epochs,
                        verbose=1,
                        use_multiprocessing=True,
                        workers=args.threads,
                        max_queue_size=args.max_queue_size,
                        callbacks=cur_callbacks)
    model.save_weights(model_output_path+".weights")
    architecture_string=model.to_json()
    outf=open(args.model_hdf5+".arch",'w')
    outf.write(architecture_string)
    print("complete!!")


def train(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args)

    w1=args.w1
    w0=args.w0

    if args.train_path==None:
        args.train_path=args.data_path
    if args.valid_path==None:
        args.valid_path=args.data_path

    if args.train_upsample==None:
        train_upsample=False
        train_upsample_ratio=0
    else:
        train_upsample=True
        train_upsample_ratio=args.train_upsample 
    train_generator=DataGenerator(args.train_path,
                                  args.ref_fasta,
                                  upsample=train_upsample,
                                  upsample_ratio=train_upsample_ratio,
                                  chroms_to_use=args.train_chroms,
                                  get_w1_w0=args.weighted)
    print("generated training data generator!")
    if args.valid_upsample==None:
        valid_upsample=False
        valid_upsample_ratio=0
    else:
        valid_upsample=True
        valid_upsample_ratio=args.valid_upsample
    valid_generator=DataGenerator(args.valid_path,
                                   args.ref_fasta,
                                   upsample=valid_upsample,
                                   upsample_ratio=valid_upsample_ratio,
                                   chroms_to_use=args.validation_chroms)
    print("generated validation data generator!")
        
    if (args.weighted==True and (w1==None or w0==None)):
        if args.w1_file==None:
            w1=train_generator.w1
            w0=train_generator.w0        
            assert args.save_w1_w0 !=None
            with open(args.save_w1_w0, 'w') as weight_file:
                weight_file.write("--w1")
                for i in w1:
                    weight_file.write(" " + str(i))
                weight_file.write(" \\" + "\n--w0")
                for i in w0:
                    weight_file.write(" " + str(i))
        else:
            w0=[float(i) for i in open(args.w0_file,'r').read().strip().split('\n')]
            w1=[float(i) for i in open(args.w1_file,'r').read().strip().split('\n')]
        print("got weights!")
    
    try:
        if (args.architecture_from_file!=None):
            architecture_module=imp.load_source('',args.architecture_from_file)
        else:
            architecture_module=importlib.import_module('kerasAC.architectures.'+args.architecture_spec)
    except:
        raise Exception("could not import requested architecture, is it installed in kerasAC/kerasAC/architectures? Is the file with the requested architecture specified correctly?")
    model=architecture_module.getModelGivenModelOptionsAndWeightInits(w0,
                                                                      w1,
                                                                      args.init_weights,
                                                                      args.from_checkpoint_weights,
                                                                      args.from_checkpoint_arch,
                                                                      args.num_tasks,args.seed)
    print("compiled the model!")
    fit_and_evaluate(model,train_generator,
                     valid_generator,args)

def main():
    args=parse_args()
    train(args)
    

if __name__=="__main__":
    main()



import sys, os
import os.path
from collections import OrderedDict
import argparse



ltrdict = {'a':[1,0,0,0],
           'c':[0,1,0,0],
           'g':[0,0,1,0],
           't':[0,0,0,1],
           'n':[0,0,0,0],
           'A':[1,0,0,0],
           'C':[0,1,0,0],
           'G':[0,0,1,0],
           'T':[0,0,0,1],
           'N':[0,0,0,0]}



class DefaultOrderedDictWrapper(object):
    def __init__(self, factory):
        self.ordered_dict = OrderedDict()
        assert hasattr(factory, '__call__')
        self.factory = factory

    def __getitem__(self, key):
        if key not in self.ordered_dict:
            self.ordered_dict[key] = self.factory() 
        return self.ordered_dict[key]



def enum(**enums):
    class Enum(object):
        pass
    to_return = Enum
    for key,val in enums.items():
        if hasattr(val, '__call__'): 
            setattr(to_return, key, staticmethod(val))
        else:
            setattr(to_return, key, val)
    to_return.vals = [x for x in enums.values()]
    to_return.the_dict = enums
    return to_return


def combine_enums(*enums):
    new_enum_dict = OrderedDict()
    for an_enum in enums:
        new_enum_dict.update(an_enum.the_dict)
    return enum(**new_enum_dict)



def args_object_from_args_dict(args_dict):
    #create an argparse.Namespace from the dictionary of inputs
    args_object=argparse.Namespace()
    #set the defaults for training/prediction/interpretation/cross-validation
    #training 
    vars(args_object)['batch_size']=1000
    vars(args_object)['num_train']=700000
    vars(args_object)['num_valid']=150000
    vars(args_object)['ref_fasta']='/mnt/data/annotations/by_release/hg19.GRCh37/hg19.genome.fa'
    vars(args_object)['init_weights']=None
    vars(args_object)['w0_file']=None
    vars(args_object)['w1_file']=None
    vars(args_object)['from_checkpoint_weights']=None
    vars(args_object)['from_checkpoint_arch']=None
    vars(args_object)['vcf_file']=None
    vars(args_object)['epochs']=40
    #prediction

    #cross-validation

    
    for key in args_dict:
        vars(args_object)[key]=args_dict[key]
        
    
    



import kerasAC.metrics
import kerasAC.activations 
from kerasAC.accuracy_metrics import *
import argparse
import yaml 
import h5py 
import pickle
import numpy as np 
import keras 
from keras.losses import *
from kerasAC.custom_losses import * 
import random
from .generators import *
from .config import args_object_from_args_dict


def get_predictions_hammock(args,model):
    import pysam
    import pandas as pd
    num_generated=0
    ref=pysam.FastaFile(args.ref_fasta) 
    data=pd.read_csv(args.data_hammock,header=None,sep='\t')
    ltrdict = {'a':[1,0,0,0],'c':[0,1,0,0],'g':[0,0,1,0],'t':[0,0,0,1], 'n':[0,0,0,0],'A':[1,0,0,0],'C':[0,1,0,0],'G':[0,0,1,0],'T':[0,0,0,1],'N':[0,0,0,0]}
    #iterate through batches and one-hot-encode on the fly
    num_entries=data.shape[0]
    predictions=None
    all_names=[] 
    while num_generated < num_entries:
        print(str(num_generated))
        start_index=num_generated
        end_index=min([num_entries,start_index+args.batch_size])
        seqs=[]
        names=[]
        for i in range(start_index,end_index):
            cur_row=data.iloc[i]
            chrom=cur_row[0]
            start_val=cur_row[1]
            end_val=cur_row[2]
            peak_metadata=cur_row[3].split(',') 
            peak_name=','.join([peak_metadata[-3],peak_metadata[-2]])
            if args.center_on_summit==True:
                summit_offset=int(peak_metadata[-1].split('[')[1].split(']')[0])
                summit_pos=start_val+summit_offset
                start_val=summit_pos - args.flank
                end_val=summit_pos+args.flank
            if start_val<1:
                start_val=1
                end_val=1+2*args.flank 
            peak_name='\t'.join([str(i) for i in [chrom,start_val,end_val,peak_name]])
            names.append(peak_name)
            try:
                seq=ref.fetch(chrom,start_val,end_val)
            except:
                continue
            seqs.append(seq)
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        if (args.squeeze_input_for_gru==False):
            #expand dimension of 1
            x=np.expand_dims(seqs,1)
        else:
            x=seqs
        try:
            predictions_batch=model.predict(x)
        except:
            print("could not get predictions -- chances are reference assembly is wrong, or bed region lies outside of chrom sizes") 
        #add the batch predictions to the full set of predictions
        if type(predictions)==type(None):
            predictions=predictions_batch
        elif type(predictions)==np.ndarray:
            predictions=np.concatenate((predictions,predictions_batch),axis=0)
            print(predictions.shape)
        elif type(predictions)==type({}):
            for key in predictions_batch:
                predictions[key]=np.concatenate((predictions[key],predictions_batch[key]),axis=0)
        else:
            print("Unsupported data type for predictions: must be np.ndarray, None, or dictionary")
            pdb.set_trace()
        all_names=all_names+names 
        num_generated+=(end_index-start_index)
        
    return [predictions,data,all_names]

def get_predictions(args,model):
    import pysam
    import pandas as pd
    test_generator=DataGenerator(args.data_path,args.ref_fasta,upsample=False,add_revcomp=False,batch_size=1000,chroms_to_use=args.predict_chroms)
    test_predictions=model.predict_generator(test_generator,
                                             max_queue_size=args.max_queue_size,
                                             workers=args.workers,
                                             use_multiprocessing=True,
                                             verbose=1)
    return [predictions,test_generator.data,None]

def get_predictions_variant(args,model):
    import pysam
    ref=pysam.FastaFile(args.ref_fasta)
    ltrdict = {'a':[1,0,0,0],'c':[0,1,0,0],'g':[0,0,1,0],'t':[0,0,0,1], 'n':[0,0,0,0],'A':[1,0,0,0],'C':[0,1,0,0],'G':[0,0,1,0],'T':[0,0,0,1],'N':[0,0,0,0]}
    #assumes the bed file has `chrom start ref alt` entries
    data=[i.split('\t') for i in open(args.variant_bed,'r').read().strip().split('\n')]
    #original
    seqs=[]
    #snp
    seqs_snp=[]
    #n bases around snp (i.e. knock out an enhancer)
    seqs_enhancer=[] 
    for entry in data[1::]:
        #introduce the specified variant 
        start_pos=int(entry[1])-(args.flank)
        end_pos=int(entry[1])+args.flank
        seq=ref.fetch(entry[0],start_pos,end_pos)
        seqs.append(seq)        
        alt_allele=entry[3]
        if alt_allele=="NA":
            #randomly insert another base
            ref_allele=seq[args.flank-1].upper()
            options=['A','C','G','T']
            options.remove(ref_allele)
            alt_allele=options[random.randint(0,2)]
        seq=seq[0:args.flank-1]+alt_allele+seq[args.flank:len(seq)]
        seqs_snp.append(seq)

        seq=np.array([ltrdict[x] for x in seq])
        start_mask=args.flank-1-args.mask
        end_mask=args.flank-1+args.mask
        seq=seq*1.0
        seq[start_mask:end_mask]=args.background_freqs
        seqs_enhancer.append(seq)
        
    seqs=np.array([[ltrdict[x] for x in seq] for seq in seqs])    
    seqs_snp=np.array([[ltrdict[x] for x in seq] for seq in seqs_snp])
    seqs_enhancer=np.asarray(seqs_enhancer)

    seqs=np.expand_dims(seqs,1)
    seqs_snp=np.expand_dims(seqs_snp,1)
    seqs_enhancer=np.expand_dims(seqs_enhancer,1)
    
    predictions_seqs=model.predict(seqs)
    predictions_seqs_snp=model.predict(seqs_snp)
    predictions_seqs_enhancer=model.predict(seqs_enhancer)
    return [predictions_seqs,predictions_seqs_snp,predictions_seqs_enhancer]

def parse_args():
    parser=argparse.ArgumentParser(description='Provide a model yaml & weights files & a dataset, get model predictions and accuracy metrics')
    parser.add_argument('--model_hdf5',help='hdf5 file that stores the model')
    parser.add_argument('--weights',help='weights file for the model')
    parser.add_argument('--yaml',help='yaml file for the model')
    parser.add_argument('--json',help='json file for the model')
    parser.add_argument('--data_hdf5',help='hdf5 file that stores the data')
    parser.add_argument('--data_path',required=True)
    parser.add_argument('--predict_chroms',default=None) 
    parser.add_argument('--data_hammock',help='input file is in hammock format, with unique id for each peak')
    parser.add_argument('--variant_bed')
    parser.add_argument('--predictions_pickle',help='name of pickle to save predictions',required=True)
    parser.add_argument('--accuracy_metrics_file',help='file name to save accuracy metrics',required=True)
    parser.add_argument('--predictions_pickle_to_load',help="if predictions have already been generated, provide a pickle with them to just compute the accuracy metrics",default=None)
    parser.add_argument('--batch_size',type=int,help='batch size to use to make model predictions',default=50)
    parser.add_argument('--sequential',default=False,help='use this flag if your model is a sequential model',action="store_true")
    parser.add_argument('--ref_fasta',default="/mnt/data/annotations/by_release/hg19.GRCh37/hg19.genome.fa")
    parser.add_argument('--background_freqs',default=None)
    parser.add_argument('--w1',nargs="*",type=float)
    parser.add_argument('--w0',nargs="*",type=float)
    parser.add_argument('--flank',default=500,type=int)
    parser.add_argument('--mask',default=10,type=int)
    parser.add_argument('--squeeze_input_for_gru',action='store_true')
    parser.add_argument('--center_on_summit',default=False,action='store_true',help="if this is set to true, the peak will be centered at the summit (must be last entry in bed file or hammock) and expanded args.flank to the left and right")
    return parser.parse_args()

def get_model(args):
    custom_objects={"positive_accuracy":kerasAC.metrics.positive_accuracy,
                    "negative_accuracy":kerasAC.metrics.negative_accuracy,
                    "precision":kerasAC.metrics.precision,
                    "recall":kerasAC.metrics.recall,
                    "softMaxAxis1":kerasAC.activations.softMaxAxis1}
    if args.w0!=None:
        w0=args.w0
        w1=args.w1
        loss_function=get_weighted_binary_crossentropy(w0,w1)
        custom_objects["weighted_binary_crossentropy"]=loss_function
    try:
        if args.yaml!=None:
            from keras.models import model_from_yaml
            #load the model architecture from yaml
            yaml_string=open(args.yaml,'r').read()
            model=model_from_yaml(yaml_string,custom_objects=custom_objects) 
            #load the model weights
            model.load_weights(args.weights)
        elif args.json!=None:
            from keras.models import model_from_json
            #load the model architecture from json
            json_string=open(args.json,'r').read()
            model=model_from_json(json_string,custom_objects=custom_objects)
            model.load_weights(args.weights)
        elif args.model_hdf5!=None: 
            #load from the hdf5
            from keras.models import load_model
            model=load_model(args.model_hdf5,custom_objects=custom_objects)
        print("got model architecture")
        print("loaded model weights")        
        
    except: 
        print("Failed to load model. HINT: if you're using weighted binary cross entropy loss, chances are you forgot to provide the --w0 or --w1 flags")
    return model

def get_predictions(args,model):
    if args.variant_bed!=None:
        predictions=get_predictions_variant(args,model)
    elif args.data_hammock!=None:
        predictions=get_predictions_hammock(args,model) 
    else:
        predictions=get_predictions(args,model) 
    print('got model predictions')
    return predictions

def predict(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args) 
    
    #get the predictions
    if args.predictions_pickle_to_load!=None:
        #load the pickled predictions
        with open(args.predictions_pickle_to_load,'rb') as handle:
            predictions=pickle.load(handle)
    else:
        #get the model
        model=get_model(args)
        predictions=get_predictions(args,model)
        
        
    if args.predictions_pickle!=None:
        #pickle the predictions in case an error occurs downstream
        #this will allow for easy recovery of model predictions without having to regenerate them
        with open(args.predictions_pickle,'wb') as handle:
            pickle.dump(predictions,handle,protocol=pickle.HIGHEST_PROTOCOL)
        print("pickled the model predictions to file:"+str(args.predictions_pickle))

    if args.accuracy_metrics_file!=None:
        print('computing accuracy metrics...')
        recallAtFDR50=dict()
        recallAtFDR20=dict()
        auroc_vals=dict()
        auprc_vals=dict()
        unbalanced_accuracy_vals=dict()
        balanced_accuracy_vals=dict()
        positives_accuracy_vals=dict()
        negatives_accuracy_vals=dict()
        num_positive_vals=dict()
        num_negative_vals=dict()
        
        outputs=np.asarray(predictions[1])
        predictions=predictions[0]
        if not(isinstance(predictions,dict)):
            predictions={'output':predictions}
        if not(isinstance(outputs,dict)):
            outputs={'output':outputs} 
        for output_mode in predictions: 
            #compute the accuracy metrics
            recallAtFDR50[output_mode]=recall_at_fdr_function(predictions[output_mode],outputs[output_mode],50)
            print('got recall at FDR50!') 
            recallAtFDR20[output_mode]=recall_at_fdr_function(predictions[output_mode],outputs[output_mode],20)
            print('got recall at FDR20!')
            auroc_vals[output_mode]=auroc_func(predictions[output_mode],outputs[output_mode])
            print('got auROC vals!')
            auprc_vals[output_mode]=auprc_func(predictions[output_mode],outputs[output_mode])
            print('got auPRC vals!')
            unbalanced_accuracy_vals[output_mode]=unbalanced_accuracy(predictions[output_mode],outputs[output_mode])
            print('got unbalanced accuracy')
            balanced_accuracy_vals[output_mode]=balanced_accuracy(predictions[output_mode],outputs[output_mode])
            print('got balanced accuracy')
            positives_accuracy_vals[output_mode]=positives_accuracy(predictions[output_mode],outputs[output_mode])
            print('got positives accuracy')
            negatives_accuracy_vals[output_mode]=negatives_accuracy(predictions[output_mode],outputs[output_mode])
            print('got negative accuracy vals')
            num_positive_vals[output_mode]=num_positives(predictions[output_mode],outputs[output_mode])
            print('got number of positive values')
            num_negative_vals[output_mode]=num_negatives(predictions[output_mode],outputs[output_mode])
            print('got number of negative values')

        #write accuracy metrics to output file: 
        print('writing accuracy metrics to file...')
        outf=open(args.accuracy_metrics_file,'w')
        for key in recallAtFDR50.keys():
            outf.write('recallAtFDR50\t'+str(key)+'\t'+'\t'.join([str(i) for i in recallAtFDR50[key]])+'\n')
        for key in recallAtFDR20.keys():
            outf.write('recallAtFDR20\t'+str(key)+'\t'+'\t'.join([str(i) for i in recallAtFDR20[key]])+'\n')
        for key in auroc_vals.keys():
            outf.write('auroc_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in auroc_vals[key]])+'\n')
        for key in auprc_vals.keys():
            outf.write('auprc_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in auprc_vals[key]])+'\n')
        for key in unbalanced_accuracy_vals.keys():
            outf.write('unbalanced_accuracy_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in unbalanced_accuracy_vals[key]])+'\n')
        for key in balanced_accuracy_vals.keys():
            outf.write('balanced_accuracy_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in balanced_accuracy_vals[key]])+'\n')
        for key in positives_accuracy_vals.keys():
            outf.write('positives_accuracy_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in positives_accuracy_vals[key]])+'\n')
        for key in negatives_accuracy_vals.keys():
            outf.write('negatives_accuracy_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in negatives_accuracy_vals[key]])+'\n')    
        for key in num_positive_vals.keys():
            outf.write('num_positive_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in num_positive_vals[key]])+'\n')
        for key in num_negative_vals.keys():
            outf.write('num_negative_vals\t'+str(key)+'\t'+'\t'.join([str(i) for i in num_negative_vals[key]])+'\n')


def main():
    args=parse_args()
    predict(args) 

    

if __name__=="__main__":
    main()
    



import argparse
import numpy as np 
import pdb
from concise.utils.plot import seqlogo, seqlogo_fig
import pysam
import argparse
import pyBigWig
import matplotlib.pyplot as plt

def parse_args():
    parser=argparse.ArgumentParser(description="generate plot of sequence importance scores")
    parser.add_argument("--input_bigwig")
    parser.add_argument("--chrom",nargs="+")
    parser.add_argument("--startpos",nargs="+",type=int)
    parser.add_argument("--endpos",nargs="+",type=int)
    parser.add_argument("--outf",default=None)
    parser.add_argument("--outf_width",type=float,default=3)
    parser.add_argument("--outf_length",type=float,default=25)
    parser.add_argument("--ref")
    parser.add_argument("--minthresh",type=float,default=None)
    parser.add_argument("--maxthresh",type=float,default=None) 
    return parser.parse_args() 
def one_hot_encode_sequence(chrom,start,end,ref):    
    num_generated=0
    ref=pysam.FastaFile(ref)
    ltrdict = {'a':[1,0,0,0],'c':[0,1,0,0],'g':[0,0,1,0],'t':[0,0,0,1], 'n':[0,0,0,0],'A':[1,0,0,0],'C':[0,1,0,0],'G':[0,0,1,0],'T':[0,0,0,1],'N':[0,0,0,0]}
    seq=ref.fetch(chrom,start,end)
    seq=np.array([ltrdict.get(x,[0,0,0,0]) for x in seq])
    return seq

def plot_seq_importance(scores, data, ylim=None, figsize=(25, 3),outf=None,tick_interval=5):
    seq_len = data.shape[0]
    product=np.expand_dims(scores,axis=1)*data
    product=np.nan_to_num(product,copy=False)
    print(product.shape)
    seqlogo_fig(product, figsize=figsize)
    plt.xticks(list(range(0, product.shape[0], tick_interval)))
    if ylim!=None:
        plt.ylim(ylim)
    if outf==None:
        plt.show()
    else:
        plt.axis("off")
        plt.gca().set_position([0,0,1,1])
        plt.savefig(outf)

def main():
    args=parse_args()
    for i in range(len(args.startpos)):
        cur_chrom=args.chrom[i]
        cur_startpos=args.startpos[i]
        cur_endpos=args.endpos[i]
        cur_data=one_hot_encode_sequence(cur_chrom,cur_startpos,cur_endpos,args.ref)
        #extract the signal from the bigwig input file
        bw=pyBigWig.open(args.input_bigwig)
        scores=bw.values(cur_chrom,cur_startpos,cur_endpos)
        if args.minthresh!=None:
            scores=[max([i,args.minthresh]) for i in scores]
        if args.maxthresh!=None:
            scores=[max([i,args.maxthresh]) for i in scores]
            
        #get the one-hot-encoded sequence in this interval
        seq=one_hot_encode_sequence(cur_chrom,cur_startpos,cur_endpos,args.ref)
        #generate the plot for the current interval
        plot_seq_importance(scores, seq, figsize=(args.outf_length, args.outf_width),outf=args.outf)

if __name__=="__main__":
    main()
    



import argparse
import pdb 
def args_object_from_args_dict(args_dict):
    #create an argparse.Namespace from the dictionary of inputs
    args_object=argparse.Namespace()
    #set the defaults for training/prediction/interpretation/cross-validation
    #training 
    vars(args_object)['batch_size']=1000
    vars(args_object)['num_train']=700000
    vars(args_object)['num_valid']=150000
    vars(args_object)['train_chroms']=None
    vars(args_object)['validation_chroms']=None 
    vars(args_object)['ref_fasta']='/mnt/data/annotations/by_release/hg19.GRCh37/hg19.genome.fa'
    vars(args_object)['init_weights']=None
    vars(args_object)['w0_file']=None
    vars(args_object)['w1_file']=None
    vars(args_object)['squeeze_input_for_gru']=False 
    vars(args_object)['from_checkpoint_weights']=None
    vars(args_object)['from_checkpoint_arch']=None
    vars(args_object)['vcf_file']=None
    vars(args_object)['epochs']=40
    vars(args_object)['patience']=3
    vars(args_object)['patience_lr']=2
    vars(args_object)['architecture_spec']="basset_architecture_multitask"
    vars(args_object)['architecture_from_file']=None
    vars(args_object)['tensorborad_logdir']='logs'
    vars(args_object)['seed']=1234
    vars(args_object)['train_path']=None
    vars(args_object)['valid_path']=None 
    vars(args_object)['train_upsample']=None
    vars(args_object)['valid_upsample']=None
    vars(args_object)['threads']=1
    vars(args_object)['max_queue_size']=100
    vars(args_object)['save_weights']=None
    vars(args_object)['w1']=None
    vars(args_object)['w0']=None    
    
    #prediction
    vars(args_object)['predict_chroms']=None
    vars(args_object)['prediction_pickle']=None
    vars(args_object)['accuracy_metrics_file']=None
    vars(args_object)['predictions_pickle_to_load']=None
    vars(args_object)['sequential']=False
    vars(args_object)['background_freqs']=None
    vars(args_object)['flank']=500
    vars(args_object)['mask']=10
    vars(args_object)['center_on_summit']=False
    
    #cross-validation
    vars(args_object)['assembly']='hg19'

    #interpret
    vars(args_object)['method']='deeplift'
    vars(args_object)['background_freqs']=None
    vars(args_object)['chromsizes']="/mnt/data/annotations/by_release/hg19.GRCh37/hg19.chrom.sizes"
    for key in args_dict:
        vars(args_object)[key]=args_dict[key]
    args=args_object
    return args

    
    



from sklearn.metrics import roc_auc_score
from sklearn.metrics import average_precision_score
import numpy as np
import kerasAC.util 
from collections import OrderedDict, defaultdict
import sys 
import pdb

def remove_ambiguous_peaks(predictions, true_y): 
    indices_to_remove = np.where(true_y < 0)
    true_y_filtered = np.delete(true_y, indices_to_remove)
    predictions_filtered = np.delete(predictions, indices_to_remove)
    return predictions_filtered, true_y_filtered


def auroc_func(predictions, true_y,thresh=None):
    [num_rows, num_cols] = true_y.shape 
    aurocs=[]
    for c in range(num_cols): 
        true_y_for_task = true_y[:,c]
        predictions_for_task = predictions[:,c]
        predictions_for_task_filtered, true_y_for_task_filtered = \
         remove_ambiguous_peaks(predictions_for_task, true_y_for_task)
        try:
            task_auroc = roc_auc_score(y_true=true_y_for_task_filtered,
                                   y_score=predictions_for_task_filtered)
        except Exception as e:
            #if there is only one class in the batch of true_y, then auROC cannot be calculated
            print("Could not calculate auROC:")
            print(str(e))
            task_auroc=None 
        aurocs.append(task_auroc) 
    return aurocs

def auprc_func(predictions, true_y,thresh=None):
    # sklearn only supports 2 classes (0,1) for the auPRC calculation
    [num_rows, num_cols]=true_y.shape 
    auprcs=[]
    for c in range(num_cols): 
        true_y_for_task=np.squeeze(true_y[:,c])
        predictions_for_task=np.squeeze(predictions[:,c])
        predictions_for_task_filtered,true_y_for_task_filtered = \
         remove_ambiguous_peaks(predictions_for_task, true_y_for_task)
        try:
            task_auprc = average_precision_score(true_y_for_task_filtered, predictions_for_task_filtered);
        except:
            print("Could not calculated auPRC:")
            print(sys.exc_info()[0])
            task_auprc=None 
        auprcs.append(task_auprc) 
    return auprcs;

def get_accuracy_stats_for_task(predictions, true_y, c):
    true_y_for_task=np.squeeze(true_y[:,c])
    predictions_for_task=np.squeeze(predictions[:,c])
    predictions_for_task_filtered,true_y_for_task_filtered = remove_ambiguous_peaks(predictions_for_task,true_y_for_task)
    predictions_for_task_filtered_round = np.array([round(el) for el in predictions_for_task_filtered])
    accuratePredictions = predictions_for_task_filtered_round==true_y_for_task_filtered;

    numPositives_forTask=np.sum(true_y_for_task_filtered==1,axis=0,dtype="float");
    numNegatives_forTask=np.sum(true_y_for_task_filtered==0,axis=0,dtype="float"); 

    accuratePredictions_positives = np.sum(accuratePredictions*(true_y_for_task_filtered==1),axis=0);
    accuratePredictions_negatives = np.sum(accuratePredictions*(true_y_for_task_filtered==0),axis=0);
    returnDict = {
        'accuratePredictions': accuratePredictions,
        'numPositives_forTask': numPositives_forTask,
        'numNegatives_forTask': numNegatives_forTask,
        'true_y_for_task_filtered': true_y_for_task_filtered,
        'predictions_for_task_filtered': predictions_for_task_filtered,
        'accuratePredictions_positives': accuratePredictions_positives,
        'accuratePredictions_negatives': accuratePredictions_negatives
    }
    return returnDict


def unbalanced_accuracy(predictions, true_y,thresh=None):
    assert predictions.shape==true_y.shape;
    assert len(predictions.shape)==2;
    [num_rows, num_cols]=true_y.shape 
    unbalanced_accuracies = []
    for c in range(num_cols):
        r = get_accuracy_stats_for_task(predictions, true_y, c)
        unbalancedAccuracy_forTask = (r['accuratePredictions_positives'] + r['accuratePredictions_negatives'])/(r['numPositives_forTask']+r['numNegatives_forTask']).astype("float");
        unbalanced_accuracies.append(unbalancedAccuracy_forTask) 
    return unbalanced_accuracies;

def positives_accuracy(predictions,true_y,thresh=None):
    assert predictions.shape==true_y.shape;
    assert len(predictions.shape)==2;
    [num_rows, num_cols]=true_y.shape 
    positive_accuracies = []
    for c in range(num_cols):
        r = get_accuracy_stats_for_task(predictions, true_y, c)
        positiveAccuracy_forTask = float(r['accuratePredictions_positives'])/float(r['numPositives_forTask'])
        positive_accuracies.append(positiveAccuracy_forTask) 
    return positive_accuracies;

def negatives_accuracy(predictions,true_y,thresh=None):
    assert predictions.shape==true_y.shape;
    assert len(predictions.shape)==2;
    [num_rows, num_cols]=true_y.shape 
    negative_accuracies = []
    for c in range(num_cols):
        r = get_accuracy_stats_for_task(predictions, true_y, c)
        negativeAccuracy_forTask = float(r['accuratePredictions_negatives'])/float(r['numNegatives_forTask'])
        negative_accuracies.append(negativeAccuracy_forTask) 
    return negative_accuracies;

def num_positives(predictions,true_y,thresh=None):
    assert predictions.shape==true_y.shape;
    assert len(predictions.shape)==2;
    [num_rows, num_cols]=true_y.shape 
    num_positives = []
    for c in range(num_cols):
        r = get_accuracy_stats_for_task(predictions, true_y, c)
        num_positives.append(r['numPositives_forTask']) 
    return num_positives;

def num_negatives(predictions,true_y,thresh=None):
    assert predictions.shape==true_y.shape;
    assert len(predictions.shape)==2;
    [num_rows, num_cols]=true_y.shape 
    num_negatives = []
    for c in range(num_cols):
        r = get_accuracy_stats_for_task(predictions, true_y, c)
        num_negatives.append(r['numNegatives_forTask'])
    return num_negatives;

def balanced_accuracy(predictions, true_y, thresh=None):
    assert predictions.shape==true_y.shape;
    assert len(predictions.shape)==2;
    [num_rows, num_cols]=true_y.shape 
    balanced_accuracies = []
    for c in range(num_cols): 
        r = get_accuracy_stats_for_task(predictions, true_y, c)
    
        positivesAccuracy_forTask = r['accuratePredictions_positives']/r['numPositives_forTask'];
        negativesAccuracy_forTask = r['accuratePredictions_negatives']/r['numNegatives_forTask'];

        balancedAccuracy_forTask= (positivesAccuracy_forTask+negativesAccuracy_forTask)/2;
        balanced_accuracies.append(balancedAccuracy_forTask) 
    return balanced_accuracies
def recall_at_fdr_function(predictions,true_y,thresh):
    if float(thresh)>1: 
        thresh=float(thresh)/100 
    [num_rows, num_cols]=true_y.shape
    recall_at_fdr_vals=[]
    for c in range(num_cols): 
        true_y_for_task=np.squeeze(true_y[:,c])
        predictions_for_task=np.squeeze(predictions[:,c])
        predictions_for_task_filtered,true_y_for_task_filtered = remove_ambiguous_peaks(predictions_for_task, true_y_for_task)

        #group by predicted prob
        predictedProbToLabels = defaultdict(list)
        for predictedProb, trueY in zip(predictions_for_task_filtered, true_y_for_task_filtered):
            predictedProbToLabels[predictedProb].append(trueY)
        #sort in ascending order of confidence
        sortedThresholds = sorted(predictedProbToLabels.keys())
        toReturnDict = OrderedDict();
        thresholdPairs=[("recallAtFDR"+str(thresh),thresh)]
        #sort desired recall thresholds by descending order of fdr        
        totalPositives = np.sum(true_y_for_task_filtered)
        totalNegatives = np.sum(1-true_y_for_task_filtered)
        #start at 100% recall
        confusionMatrixStatsSoFar = [[0,totalNegatives]
                                    ,[0,totalPositives]]
        recallsForThresholds = []; #for debugging
        fdrsForThresholds = [];

        for threshold in sortedThresholds:
            labelsAtThreshold=predictedProbToLabels[threshold];
            positivesAtThreshold=sum(labelsAtThreshold)
            negativesAtThreshold = len(labelsAtThreshold)-positivesAtThreshold
            
            #when you cross this threshold they all get predicted as negatives.
            confusionMatrixStatsSoFar[0][0] += negativesAtThreshold
            confusionMatrixStatsSoFar[0][1] -= negativesAtThreshold
            confusionMatrixStatsSoFar[1][0] += positivesAtThreshold
            confusionMatrixStatsSoFar[1][1] -= positivesAtThreshold
            totalPredictedPositives = confusionMatrixStatsSoFar[0][1]\
                                  + confusionMatrixStatsSoFar[1][1]
            fdr = 1 - (confusionMatrixStatsSoFar[1][1]/
                       float(totalPredictedPositives))\
                       if totalPredictedPositives > 0 else 0.0
            recall = confusionMatrixStatsSoFar[1][1]/float(totalPositives)
            recallsForThresholds.append(recall)
            fdrsForThresholds.append(fdr)
            #first index of a thresholdPair is the name, second idx
            #is the actual threshold
            while (len(thresholdPairs)>0 and fdr<=thresholdPairs[0][1]):
                toReturnDict[thresholdPairs[0][0]]=recall
                thresholdPairs=thresholdPairs[1::]
            if len(thresholdPairs)==0:
                break;
        for thresholdPair in thresholdPairs:
            toReturnDict[thresholdPairs[0][0]]=0.0
        recall_at_fdr_vals.append(toReturnDict['recallAtFDR'+str(thresh)])
    return recall_at_fdr_vals


AccuracyStats = kerasAC.util.enum(
    auROC="auROC",
    auPRC="auPRC",
    balanced_accuracy="balanced_accuracy",
    unbalanced_accuracy="unbalanced_accuracy",
    positives_accuracy="positives_accuracy",
    negatives_accuracy="negatives_accuracy",
    num_positives="num_positives",
    num_negatives="num_negatives",
    onehot_rows_crossent="onehot_rows_crossent",
    recall_at_fdr50="recallAtFDR50",
    recall_at_fdr20="recallAtFDR20",
    recall_at_fdr10="recallAtFDR10",
    recall_at_fdr1="recallAtFDR1",
    recall_at_fdr="recallAtFDR")

compute_func_lookup = {
    AccuracyStats.auROC: auroc_func,
    AccuracyStats.auPRC: auprc_func,
    AccuracyStats.balanced_accuracy: balanced_accuracy,
    AccuracyStats.unbalanced_accuracy: unbalanced_accuracy,
    AccuracyStats.recall_at_fdr50: recall_at_fdr_function,
    AccuracyStats.recall_at_fdr20: recall_at_fdr_function,
    AccuracyStats.recall_at_fdr10: recall_at_fdr_function,
    AccuracyStats.recall_at_fdr1: recall_at_fdr_function,
    AccuracyStats.recall_at_fdr: recall_at_fdr_function,
    AccuracyStats.positives_accuracy:positives_accuracy,
    AccuracyStats.negatives_accuracy:negatives_accuracy,
    AccuracyStats.num_positives:num_positives,
    AccuracyStats.num_negatives:num_negatives
}
is_larger_better_lookup = {
    AccuracyStats.auROC: True,
    AccuracyStats.auPRC: True,
    AccuracyStats.balanced_accuracy: True,
    AccuracyStats.unbalanced_accuracy: True,
    AccuracyStats.recall_at_fdr50: True,
    AccuracyStats.recall_at_fdr20: True,
    AccuracyStats.recall_at_fdr10: True,
    AccuracyStats.recall_at_fdr1: True,
    AccuracyStats.recall_at_fdr: True
}

multi_level_metrics=["recallAtFDR"]




#Note: this is ugly w/ use of tf & K --> needed to avoid custom keras modifications 
import tensorflow as tf
import keras.backend as K
def spearman_corr(y_true,y_pred):
    import K.contribs.metrics.streaming_pearson_correlation
    return K.contribs.metrics.streaming_pearson_correlation(y_pred,y_true)

def positive_accuracy(y_true,y_pred):
    one_indices=tf.cast(tf.where(tf.equal(y_true,1.0)),'int32')
    y_true_subset=tf.gather_nd(y_true,one_indices)
    y_pred_subset=tf.gather_nd(y_pred,one_indices)
    positive_accuracy=K.mean(tf.equal(y_true_subset,tf.round(y_pred_subset)))
    return positive_accuracy


def negative_accuracy(y_true,y_pred):
    one_indices=tf.cast(tf.where(tf.equal(y_true,0.0)),'int32')
    y_true_subset=tf.gather_nd(y_true,one_indices)
    y_pred_subset=tf.gather_nd(y_pred,one_indices)
    negative_accuracy=K.mean(tf.equal(y_true_subset,tf.round(y_pred_subset)))
    return negative_accuracy
                                        


def precision(y_true, y_pred):
    '''Calculates the precision, a metric for multi-label classification of
    how many selected items are relevant.
    '''
    true_positives = K.sum(tf.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives =K.sum(tf.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision


def recall(y_true, y_pred):
    '''Calculates the recall, a metric for multi-label classification of
    how many relevant items are selected.
    '''
    true_positives = K.sum(tf.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(tf.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall
                                        






import numpy as np
import pysam
import pandas as pd
import pdb
import tabix

def revcomp(seq):
    seq=seq[::-1].upper()
    comp_dict=dict()
    comp_dict['A']='T'
    comp_dict['T']='A'
    comp_dict['C']='G'
    comp_dict['G']='C'
    rc=[]
    for base in seq:
        if base in comp_dict:
            rc.append(comp_dict[base])
        else:
            rc.append(base)
    return ''.join(rc)

def add_variants(bed_entries,vcf,args,ltrdict):
    for seq_index in range(len(bed_entries)):
        bed_entry=bed_entries[seq_index]
        cur_start=bed_entry[1]
        cur_end=bed_entry[2]
        max_offset=cur_end-cur_start
        try:
            variants=[i for i in vcf.query(bed_entry[0],bed_entry[1],bed_entry[2])]
        except:
            continue
        if variants!=[]:
            for variant in variants:
                if args.global_vcf == False:
                    #the vcf contains personal variants for an individual
                    pos=float(variant[1])
                    offset=int(pos-cur_start)
                    if offset >= max_offset:
                        continue
                    alleles=variant[3].split(',')+variant[4].split(',')
                    geno=[int(i) for i in variant[9].split(':')[0].split('/')]
                    try:
                        freqs1=np.asarray(ltrdict[alleles[geno[0]]])
                        freqs2=np.asarray(ltrdict[alleles[geno[1]]])
                    except:
                        continue
                    freqs=np.mean((freqs1,freqs2),axis=0)

                else:
                    #common allele freqs for the reference genome
                    pos=float(variant[1])
                    offset=int(pos-cur_start)
                    if offset >= max_offset:
                        continue
                    alleles=[variant[3],variant[4][0]]
                    try:
                        all_freqs=[float(variant[5]),float(variant[6])]
                    except:
                        continue
                    freqs=np.asarray(ltrdict[alleles[0]])*all_freqs[0]+np.asarray(ltrdict[alleles[1]])*all_freqs[1]
                seqs[seq_index][offset]=freqs
    return seqs

#currently we have on-the-fly batch generation from hdf5 & bed files.
def data_generator(data_path,args,upsample_ratio):
    if data_path.endswith('.hdf5'):
        return data_generator_hdf5(data_path,args)
    elif (data_path.endswith('.bed') or data_path.endswith('.bed.gz') or data_path.endswith('.tsv') or data_path.endswith('.tsv.gz')):
      if upsample_ratio <= 0.0:
        return data_generator_bed_original(data_path, args)
      else:
        return data_generator_bed_upsample(data_path, args, upsample_ratio)
    else:
        raise Exception("data for generator must be in hdf5 format (.hdf5 0ending) or bed format (.bed ending). Neither is true. Exiting")

def data_generator_bed_upsample(bed_source, args, upsample_ratio):
    #open the reference file
    ref=pysam.FastaFile(args.ref_fasta)
    #load the train data as a pandas dataframe, skip the header
    data=pd.read_csv(bed_source,header=0,sep='\t',index_col=[0,1,2])
    ones = data.loc[(data > 0).any(axis=1)]
    zeros = data.loc[(data < 1).all(axis=1)]
    #decide if reverse complement should be used
    if args.revcomp==True:
        batch_size=args.batch_size/2
    else:
        batch_size=args.batch_size
    pos_batch_size = int(args.batch_size * upsample_ratio)
    neg_batch_size = args.batch_size - pos_batch_size
    ltrdict = {'a':[1,0,0,0],
               'c':[0,1,0,0],
               'g':[0,0,1,0],
               't':[0,0,0,1],
               'n':[0,0,0,0],
               'A':[1,0,0,0],
               'C':[0,1,0,0],
               'G':[0,0,1,0],
               'T':[0,0,0,1],
               'N':[0,0,0,0]}
    #if vcf file is provided, load the subject's variants
    vcf=None
    if (args.vcf_file!=None):
        vcf=tabix.open(args.vcf_file)
    #iterate through batches and one-hot-encode on the fly
    pos_start_index = 0
    pos_num_generated = 0
    pos_total_entries = ones.shape[0] - pos_batch_size
    neg_start_index = 0
    neg_num_generated = 0
    neg_total_entries = zeros.shape[0] - neg_batch_size
    while True:
        if (pos_num_generated >= pos_total_entries):
            pos_start_index=0
            ones = pd.concat([ones[pos_num_generated:], ones[:pos_num_generated]])
            pos_num_generated = 0
        if (neg_num_generated >= neg_total_entries):
            neg_start_index = 0
            zeros = pd.concat([zeros[neg_num_generated:], zeros[:neg_num_generated]])
            neg_num_generated = 0
        pos_end_index = pos_start_index + int(pos_batch_size)
        neg_end_index = neg_start_index + int(neg_batch_size)
        #get seq positions
        pos_bed_entries=[(ones.index[i]) for i in range(pos_start_index,pos_end_index)]
        neg_bed_entries=[(zeros.index[i]) for i in range(neg_start_index, neg_end_index)]
        bed_entries = pos_bed_entries + neg_bed_entries
        #get sequences
        seqs=[ref.fetch(i[0],i[1],i[2]) for i in bed_entries]
        if args.revcomp==True:
            #add in the reverse-complemented sequences for training.
            seqs_rc=[revcomp(s) for s in seqs]
            seqs=seqs+seqs_rc
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        #add in subject-specific allele frequencies, if provided
        if vcf!=None:
            seqs=add_variants(bed_entries,vcf,args,ltrdict)
        #expand dimension of 1,  unless we're dealing with a GRU in recurrent network
        if(args.squeeze_input_for_gru==False):
            x_batch=np.expand_dims(seqs,1)
        else:
            x_batch=seqs
        y_labels_ones = ones[pos_start_index:pos_end_index]
        y_labels_zeros = zeros[neg_start_index:neg_end_index]
        y_batch_ones = np.asarray(y_labels_ones)
        y_batch_zeros = np.asarray(y_labels_zeros)
        y_batch = np.concatenate([y_batch_ones, y_batch_zeros])
        if args.revcomp==True:
            y_batch=np.concatenate((y_batch,y_batch),axis=0)
        pos_num_generated += pos_batch_size
        neg_num_generated += neg_batch_size
        pos_start_index = pos_end_index
        neg_start_index = neg_end_index
        if (args.squeeze_input_for_gru==False):
            if ((x_batch.ndim < 4) or (y_batch.ndim <2)):
                print("skipping!: hint-- is your reference (i.e. hg19) correct?")
                continue
            else:
                yield tuple([x_batch,y_batch])
        else:
            if ((x_batch.ndim <3) or (y_batch.ndim <2)):
                print("skipping!: hint -- is your reference (i.e. hg19) correct?")
                continue
            else:
                yield tuple([x_batch,y_batch])

def data_generator_bed_original(bed_source,args):
    #open the reference file
    ref=pysam.FastaFile(args.ref_fasta)
    #load the train data as a pandas dataframe, skip the header
    data=pd.read_csv(bed_source,header=0,sep='\t',index_col=[0,1,2])
    ltrdict = {'a':[1,0,0,0],
               'c':[0,1,0,0],
               'g':[0,0,1,0],
               't':[0,0,0,1],
               'n':[0,0,0,0],
               'A':[1,0,0,0],
               'C':[0,1,0,0],
               'G':[0,0,1,0],
               'T':[0,0,0,1],
               'N':[0,0,0,0]}
    #if vcf file is provided, load the subject's variants
    vcf=None
    if (args.vcf_file!=None):
        vcf=tabix.open(args.vcf_file)
    #iterate through batches and one-hot-encode on the fly
    start_index=0
    num_generated=0
    total_entries=data.shape[0]-args.batch_size
    #decide if reverse complement should be used
    if args.revcomp==True:
        batch_size=args.batch_size/2
    else:
        batch_size=args.batch_size
    while True:
        if (num_generated >=total_entries):
            start_index=0
        end_index=start_index+int(batch_size)
        #get seq positions
        bed_entries=[(data.index[i]) for i in range(start_index,end_index)]
        #get sequences
        seqs=[ref.fetch(i[0],i[1],i[2]) for i in bed_entries]
        if args.revcomp==True:
            #add in the reverse-complemented sequences for training.
            seqs_rc=[revcomp(s) for s in seqs]
            seqs=seqs+seqs_rc
        seqs=np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])
        #add in subject-specific allele frequencies, if provided
        if vcf!=None:
            seqs=add_variants(bed_entries,vcf,args,ltrdict)
        #expand dimension of 1,  unless we're dealing with a GRU in recurrent network
        if(args.squeeze_input_for_gru==False):
            x_batch=np.expand_dims(seqs,1)
        else:
            x_batch=seqs
        y_batch=np.asarray(data[start_index:end_index])
        if args.revcomp==True:
            y_batch=np.concatenate((y_batch,y_batch),axis=0)
        num_generated+=batch_size
        start_index=end_index
        if (args.squeeze_input_for_gru==False):
            if ((x_batch.ndim < 4) or (y_batch.ndim <2)):
                #pdb.set_trace() 
                print("skipping!: hint-- is your reference (i.e. hg19) correct?")
                continue
            else:
                yield tuple([x_batch,y_batch])
        else:
            if ((x_batch.ndim <3) or (y_batch.ndim <2)):
                print("skipping!: hint -- is your reference (i.e. hg19) correct?")
                continue
            else:
                yield tuple([x_batch,y_batch])

def data_generator_hdf5(data_path,args):
    hdf5_source=h5py.File(data_path,'r')
    num_generated=0
    total_entries=hdf5_source['X']['default_input_mode_name'].shape[0]
    start_index=0
    batch_size=args.batch_size
    while True:
        if(num_generated >=total_entries):
            start_index=0
        end_index=start_index+batch_size
        x_batch=hdf5_source['X']['default_input_mode_name'][start_index:end_index]
        y_batch=hdf5_source['Y']['default_output_mode_name'][start_index:end_index]
        num_generated+=batch_size
        start_index=end_index
        yield tuple([x_batch,y_batch])
